# osh
# Copyright (C) Jack Orenstein <jao@geophile.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

"""C{installosh [-d INSTALL_DIRECTORY] [OSH_CONFIG_FILE]}

Installs osh on a cluster. The cluster is identified using
remote execution syntax, for example::

    osh @fred [ installosh ]

If C{-d} is not specified, then
installation is to the standard site-packages directory
(e.g. /usr/lib/python2.4/site-packages). Otherwise, installation goes
to C{INSTALL_DIRECTORY}. (C{INSTALL_DIRECTORY} replaces the entire path, not just
the C{/usr} or C{/usr/lib} part of the path.)  On each node of the
cluster, a C{.oshrc} files is created in the home directory of the
user specified for C{CLUSTER}. This file is copied from
C{OSH_CONFIG_FILE}, if specified. Otherwise, C{~/.oshrc} is copied.

Not available through the API.
"""

import os
import sys
import threading

import osh.core
import osh.config
import osh.cluster
import osh.util

import progtrack

ssh = osh.util.ssh
scp = osh.util.scp

_DISTRIBUTION_FILENAME = 'osh-' + osh.config.OSH_VERSION + '.tar.gz'
_CONFIG_FILE = '.oshrc'
_PACKAGE_HOME = '/usr/share/osh'
_REMOTE_INSTALL = 'remoteinstall.py'

# CLI
def _installosh():
    return _InstallOsh()

class _InstallOsh(osh.core.RunLocal):

    _install_dir = None
    _config_file = None

    # object interface

    def __init__(self):
        osh.core.RunLocal.__init__(self, 'd:', (0, 1))    

    # BaseOp interface
    
    def doc(self):
        return __doc__

    def create_command_state(self, oshthreads):
        ui = progtrack.ProgressTrackingUI('installosh')
        ui.add_column('host', 25)
        ui.add_column(['create', 'osh dir'], 10)
        ui.add_column(['copy', 'package'], 10)
        ui.add_column(['copy', 'installer'], 10)
        ui.add_column(['run', 'installer'], 10)
        ui.add_column(['copy', 'conf file'], 10)
        for thread in oshthreads:
            host = thread.state
            ui.add_row(host.name)
        return ui

    def setup(self):
        args = self.args()
        self._install_dir = args.string_arg('-d')
        if args.has_next():
            self._config_file = args.next_string()
        else:
            self._config_file = '%s/%s' % (os.environ['HOME'], _CONFIG_FILE)
            
    # Generator interface

    def execute(self):
        # UI object actually starts only on first call to start, ignores subsequent calls.
        self.ui.start() 
        try:
            self.install()
        finally:
            self.ui.stop()


    # internals

    ui = property(lambda self: self.command_state())

    host = property(lambda self: self.thread_state)

    def install(self):
        host = self.host
        ui = self.ui
        stage = 0
        try:
            # Remove existing /usr/share/osh directory
            stage += 1
            ssh(host.user,
                host.address,
                '"rm -rf %s"' % _PACKAGE_HOME)
            # Create directory /usr/share/osh
            ssh(host.user,
                host.address,
                '"mkdir %s"' % _PACKAGE_HOME)
            ui.ok(host.name, stage)
            # Copy package to /usr/share/osh
            stage += 1
            scp(host.user,
                host.address,
                None,
                '%s/%s' % (_PACKAGE_HOME, _DISTRIBUTION_FILENAME),
                _PACKAGE_HOME)
            ui.ok(host.name, stage)
            # Copy remoteinstall script to /usr/share/osh
            stage += 1
            scp(host.user,
                host.address,
                None,
                '%s/%s' % (_PACKAGE_HOME, _REMOTE_INSTALL),
                _PACKAGE_HOME)
            ui.ok(host.name, stage)
            # Run remoteinstall script
            stage += 1
            if self._install_dir:
                ssh(host.user,
                    host.address,
                     '"%s/%s %s %s"' % (_PACKAGE_HOME, _REMOTE_INSTALL, _PACKAGE_HOME, self._install_dir))
            else:
                ssh(host.user,
                    host.address,
                    '"%s/%s %s"' % (_PACKAGE_HOME, _REMOTE_INSTALL, _PACKAGE_HOME))
            ui.ok(host.name, stage)
            # Copy config file to home directory unless one already exists
            stage += 1
            ls_oshrc, errors = ssh(host.user,
                                   host.address,
                                   '"ls /%s/%s"' % (host.user, _CONFIG_FILE))
            if len(ls_oshrc) == 0:
                scp(host.user,
                    host.address,
                    None,
                    self._config_file,
                    '~')
            ui.ok(host.name, stage)
        except:
            (exc_type, exc_value, exc_traceback) = sys.exc_info()
            message = str(exc_value).strip()
            ui.error(host.name, stage, message)

